import functools
import logging
import json
import time

import requests

from . import client



def ava_log(func):
    """
    Function decorator for implementation of logging
    @param func: function where logging needs to be implemented
    @return: function
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        logger = logging.getLogger()
        is_log_req_resp_allowed = False
        ava_log_entry = {}
        is_error_log = False

        # Logic :
        # 1) Check if decorator is called for function in AvaTaxClient class
        # 2) If logger is set then use it else use new instance of logger (this property is useful when SDK consumer
        #    wants to use specific configuration)
        # 3) Execute actual method and create log entry in case of http response only
        # 4) In case of exception though log error
        execution_start_time = time.perf_counter()
        try:
            if isinstance(args[0], client.AvataxClient):
                if hasattr(args[0], "logger") and (args[0].__getattribute__("logger") is not None):
                    logger = args[0].__getattribute__("logger")
                if hasattr(args[0], "is_log_req_resp_allowed"):
                    is_log_req_resp_allowed = args[0].__getattribute__("is_log_req_resp_allowed")
            result = func(*args, **kwargs)
            if result is not None and isinstance(result, requests.models.Response):
                ava_log_entry = get_ava_log_entry(result, is_log_req_resp_allowed)
            return result
        except Exception as e:
            is_error_log = True
            ava_log_entry["error"] = str(e)
            raise e
        finally:
            total_execution_time = time.perf_counter() - execution_start_time
            if "execution_time" not in ava_log_entry:
                ava_log_entry["execution_time"] = total_execution_time * 1000
            json_data = json.dumps(ava_log_entry, indent=4)
            if is_error_log:
                logger.error(json_data)
            else:
                logger.info(json_data)

    return wrapper


def get_ava_log_entry(result: requests.Response, is_log_req_resp_allowed: bool) -> dict:
    log_entry = {}
    log_entry["execution_time"] = result.elapsed.total_seconds() * 1000
    log_entry["x-correlation-id"] = result.headers["x-correlation-id"]
    log_entry["status_code"] = result.status_code
    log_entry["request_url"] = result.url
    log_entry["method"] = result.request.method
    if is_log_req_resp_allowed:
        log_entry["request"] = str(result.request.body)
        log_entry["response"] = result.text

    return log_entry


def decorate_all_methods(decorator, exclude=["__init__", "add_credentials"]):
    """
    Decorator to be applied at class level. This decorator decorates all methods in class with
    supplied decorator in parameter.
    @param decorator: decorator to be applied to all methods in class
    @param exclude: list of methods to exclude from being decorated.
    @return: class with decorated methods
    """

    def decorate(cls):
        for attr in cls.__dict__:
            if callable(getattr(cls, attr)) and attr not in exclude:
                setattr(cls, attr, decorator(getattr(cls, attr)))
        return cls

    return decorate



"""Test module for Create Transaction method."""
import pytest
from requests import Response


def test_201_response_when_creating_transaction(auth_client, tax_document):
    """Test recieving 200 status code when valid data passed in."""
    r = auth_client.create_transaction(tax_document, None)
    assert r.status_code == 201


def test_response_is_proper_response_object(auth_client, tax_document):
    """Test recieving 200 status code when valid data passed in."""
    r = auth_client.create_transaction(tax_document, None)
    assert isinstance(r, Response)


def test_transaction_is_saved_when_commit_is_false(auth_client, tax_document):
    """Test that a transaction is saved when commit is false."""
    tax_document['commit'] = False
    r = auth_client.create_transaction(tax_document, None)
    assert '"status":"Saved"' in r.text


def test_transaction_is_commited_when_commit_is_true(
    auth_client, tax_document
):
    """Test that a transaction is commited when commit is true."""
    tax_document['commit'] = True
    r = auth_client.create_transaction(tax_document, None)
    assert '"status":"Committed"' in r.text


def test_include_param_is_summary_only(auth_client, tax_document):
    """Test param is 'summary only' return corresponding response format."""
    include = {'$include': 'SummaryOnly'}
    r = auth_client.create_transaction(tax_document, include)
    assert 'addresses' not in r.text


def test_include_param_is_addresses(auth_client, tax_document):
    """Test param is 'addresses' return corresponding response format."""
    include = {'$include': 'Addresses'}
    r = auth_client.create_transaction(tax_document, include)
    assert 'addresses' in r.text



"""Test Transactin Builder class."""
from src.avalara.transaction_builder import TransactionBuilder


def test_initialize_builder_object(auth_client):
    """Test if a new transaction can be initialize."""
    new_trans = TransactionBuilder(
        auth_client, 'DEFAULT', 'SalesInvoice', 'ABC123'
    )
    assert isinstance(new_trans, TransactionBuilder)


def test_with_commit_method(mt_trans):
    """Test method functionality of the Transaction Builder."""
    trans = mt_trans.with_commit()
    assert trans.create_model['commit'] is True


def test_with_transaction_code_method(mt_trans):
    """Test method functionality of the Transaction Builder."""
    trans = mt_trans.with_transaction_code('1234567')
    assert trans.create_model['code'] == '1234567'


def test_with_type_method(mt_trans):
    """Test method functionality of the Transaction Builder."""
    trans = mt_trans.with_type('SalesInvoice')
    assert trans.create_model['type'] == 'SalesInvoice'


def test_with_address_method(mt_trans, valid_address):
    """Test method functionality of the Transaction Builder."""
    trans = mt_trans.with_address('SingleLocation', valid_address)
    assert trans.create_model['addresses']['SingleLocation'] == valid_address


def test_with_line_address_method(mt_trans, valid_address):
    """Test method functionality of the Transaction Builder."""
    mt_trans.with_line(20, 100, 'ITEM2001', 1234567)
    mt_trans.with_line_address('SingleLocation', valid_address)
    last_line = mt_trans.get_most_recent_line()
    assert last_line['addresses']['SingleLocation'] == valid_address


def test_with_latlong_method(mt_trans):
    """Test method functionality of the Transaction Builder."""
    lat_long = {'latitude': 47.6187, 'longitude': -122.3524}
    trans = mt_trans.with_latlong('SingleLocation', 47.6187, -122.3524)
    assert trans.create_model['addresses']['SingleLocation'] == lat_long


def test_with_line_method(mt_trans):
    """Test method functionality of the Transaction Builder."""
    temp = {
        'amount': 20,
        'number': '1',
        'quantity': 100,
        'itemCode': 'ITEM2001',
        'taxCode': '1234567',
        'taxIncluded': False,
        'exemptionCode': '',
    }
    trans = mt_trans.with_line(20, 100, 'ITEM2001', 1234567)
    assert trans.create_model['lines'][-1] == temp


def test_with_line_method_with_string_line_number(mt_trans):
    """Test method functionality of the Transaction Builder."""
    expected_resp_str_line_number = {
        'amount': 20,
        'number': 'shipping-1',
        'quantity': 100,
        'itemCode': 'ITEM2001',
        'taxCode': '1234567',
        'taxIncluded': False,
        'exemptionCode': '',
    }
    expected_resp_int_line_number = {
        'amount': 100,
        'number': '5',
        'quantity': 1,
        'itemCode': 'ITEM2005',
        'taxCode': '5555555',
        'taxIncluded': False,
        'exemptionCode': '',
    }
    trans = mt_trans.with_line(
        amount=20,
        quantity=100,
        item_code='ITEM2001',
        tax_code=1234567,
        tax_included=False,
        line_number='shipping-1',
        exemption_code="",
    ).with_line(
        amount=100,
        quantity=1,
        item_code='ITEM2005',
        tax_code=5555555,
        tax_included=False,
        line_number=5,
        exemption_code="",
    )
    assert trans.create_model['lines'][0] == expected_resp_str_line_number
    assert trans.create_model['lines'][1] == expected_resp_int_line_number


def test_with_line_method_with_optional_parameters(mt_trans):
    """Test method functionality of the Transaction Builder."""
    expected_resp_int_line_number = {
        'amount': 100,
        'number': '5',
        'quantity': 1,
        'itemCode': 'ITEM2005',
        'taxCode': '',
        'taxIncluded': True,
        'exemptionCode': 'EXEMPT_CUST_002',
    }
    trans = mt_trans.with_line(
        amount=100,
        quantity=1,
        item_code='ITEM2005',
        tax_included=True,
        line_number=5,
        exemption_code="EXEMPT_CUST_002",
    )
    assert trans.create_model['lines'][0] == expected_resp_int_line_number


def test_with_line_method_line_number_not_null(mt_trans):
    """Test method functionality of the Transaction Builder."""
    temp = {
        'amount': 20,
        'number': '2',
        'quantity': 100,
        'itemCode': 'ITEM2001',
        'taxCode': '1234567',
        'taxIncluded': False,
        'exemptionCode': '',
    }
    trans = mt_trans.with_line(20, 100, 'ITEM2001', 1234567, 2)
    assert trans.create_model['lines'][-1] == temp


def test_with_exempt_line_method(mt_trans):
    """Test method functionality of the Transaction Builder."""
    temp = {
        'amount': 20,
        'number': '1',
        'quantity': 1,
        'itemCode': 'ITEM2001',
        'exemptionCode': '1234567',
        'taxIncluded': False,
        'taxCode': '',
    }
    trans = mt_trans.with_exempt_line(20, 'ITEM2001', 1234567)
    assert trans.create_model['lines'][-1] == temp


def test_with_exempt_line_method_quantity(mt_trans):
    """Test method functionality of the Transaction Builder."""
    temp = {
        'amount': 100,
        'number': '1',
        'quantity': 5,
        'itemCode': 'ITEM2001',
        'exemptionCode': 'EXEMPT_THIS_CUSTOMER',
        'taxIncluded': False,
        'taxCode': '',
    }
    trans = mt_trans.with_exempt_line(
        quantity=5,
        amount=100,
        item_code='ITEM2001',
        exemption_code='EXEMPT_THIS_CUSTOMER',
    )
    assert trans.create_model['lines'][-1] == temp


def test_with_diagnostics_method(mt_trans):
    """Test method functionality of the Transaction Builder."""
    trans = mt_trans.with_diagnostics()
    assert trans.create_model['debugLevel'] == 'Diagnostic'


def test_with_discount_amount_method(mt_trans):
    """Test method functionality of the Transaction Builder."""
    trans = mt_trans.with_discount_amount(200)
    assert trans.create_model['discount'] == 200


def test_with_parameter_method(mt_trans):
    """Test method functionality of the Transaction Builder."""
    trans = mt_trans.with_parameter('param', 'param_val')
    assert trans.create_model['parameters']['param'] == 'param_val'


def test_with_line_parameter_method(mt_trans):
    """Test method functionality of the Transaction Builder."""
    trans = mt_trans.with_parameter('param', 'param_val')
    assert trans.create_model['parameters']['param'] == 'param_val'


def test_get_most_recent_line_method(mt_trans):
    """Test method functionality of the Transaction Builder."""
    mt_trans.with_line(20, 100, 'ITEM2001', 1234567)
    line = mt_trans.get_most_recent_line('TestName')
    assert line is mt_trans.create_model['lines'][-1]


def test_with_line_tax_overrride_method(mt_trans):
    """Test method functionality of the Transaction Builder."""
    model = {
        "type": "TaxAmount",
        "taxAmount": 6.25,
        "taxDate": "2017-12-14",
        "reason": "Precalculated Tax",
    }
    trans = mt_trans.with_tax_override(
        'TaxAmount', 'Precalculated Tax', 6.25, '2017-12-14'
    )
    assert trans.create_model['taxOverride'] == model


def test_with_separate_address_method(mt_trans, valid_address):
    """Test method functionality of the Transaction Builder."""
    model = {
        'number': 1,
        'quantity': 1,
        'amount': 200,
        'addresses': {'SingleLocation': valid_address},
    }
    trans = mt_trans.with_separate_address_line(
        200, 'SingleLocation', valid_address
    )
    assert trans.create_model['lines'][-1] == model


def test_create_adjustment_request_method(mt_trans):
    """Test method functionality of the Transaction Builder."""
    model = {
        'newTransaction': mt_trans.create_model,
        'adjustmentDescription': 'EOY adjustment',
        'adjustmentReason': 'internal transfer',
    }
    re = mt_trans.create_adjustment_request(
        'EOY adjustment', 'internal transfer'
    )
    assert re == model


def test_create_trans_without_commit(mt_trans, valid_address):
    """Test creating a transaction with transaction builder."""
    trans = (
        mt_trans.with_address('SingleLocation', valid_address)
        .with_line(20, 100, 'ITEM2001', 1234567)
        .create()
    )
    assert '"status":"Saved"' in trans.text


def test_create_trans_with_commit(mt_trans, valid_address):
    """Test creating a transaction with transaction builder."""
    trans = (
        mt_trans.with_address('SingleLocation', valid_address)
        .with_line(20, 100, 'ITEM2001', 1234567)
        .with_commit()
        .create()
    )
    assert '"status":"Committed"' in trans.text


def test_create_trans_with_two_lines(
    mt_trans, valid_address, ship_from_address
):
    """Test creating transaction with 2 lines."""
    trans = (
        mt_trans.with_address('ShipFrom', ship_from_address)
        .with_address('ShipTo', valid_address)
        .with_line(100.0, 1, None, 'P0000000')
        .create()
    )
    assert trans.status_code == 201


def test_create_trans_with_exempt_line(mt_trans, valid_address):
    """Test creating a transaction with transaction builder."""
    trans = (
        mt_trans.with_address('SingleLocation', valid_address)
        .with_line(100.0, 1, None, 'P0000000')
        .with_exempt_line(50.0, None, 'NT')
        .create()
    )
    assert trans.status_code == 201


def test_create_trans_with_line_address(
    mt_trans, valid_address, ship_from_address
):
    """Test creating a transaction with transaction builder."""
    trans = (
        mt_trans.with_line(100.0, 1, None, 'P0000000')
        .with_line_address('ShipFrom', ship_from_address)
        .with_line_address('ShipTo', valid_address)
        .create()
    )
    assert trans.status_code == 201



"""Test resolve address method."""


def test_address_fixture_import(valid_address):
    """Test address fixture imports without errors."""
    assert isinstance(valid_address, dict)


def test_resolve_address_response_code_200(auth_client, valid_address):
    """Testing client without authorization."""
    assert auth_client.resolve_address(valid_address).status_code == 200


def test_resolved_address_response_has_coordinates(auth_client, valid_address):
    """Test the response has latitude for resolved address."""
    response_json = auth_client.resolve_address(valid_address).json()
    assert isinstance(response_json['coordinates']['latitude'], float)


def test_resolve_address_response_code_400(auth_client):
    """Test resolve address with incomplete data."""
    bad_address = {'line1': 'Incomplete Address'}
    assert auth_client.resolve_address(bad_address).status_code == 400


def test_incomplete_resolve_address_response_code_401(unauth_client):
    """Test resolve address with incomplete data and unauthorized user."""
    bad_address = {'line1': 'Incomplete Address'}
    assert unauth_client.resolve_address(bad_address).status_code == 401


def test_unauthorized_resolve_address_response_code_401(unauth_client, valid_address):
    """Test resolve address with correct data and unauthorized user."""
    assert unauth_client.resolve_address(valid_address).status_code == 401


def test_resolve_address_error_message(auth_client):
    """Test that incomplete address returns error message."""
    bad_address = {'line1': 'Incomplete Address'}
    assert auth_client.resolve_address(bad_address).json()['error']



"""Test the ping method."""


def test_ping_makes_connection(unauth_client):
    """Testing client without authorization."""
    assert unauth_client.ping().status_code == 200


def test_ping_auth_is_false_connection(unauth_client):
    """Testing client without authorization."""
    assert '"authenticated":false' in unauth_client.ping().text


def test_ping_auth_is_true_using_client_id(auth_client):
    """Testing client with authorization from username and password."""
    assert '"authenticated":true' in auth_client.ping().text


def test_ping_auth_with_invalid_user_and_pass(unauth_client):
    """Testing client without proper username and password."""
    unauth_client.add_credentials('1234', '4321')
    assert '"authenticated":false' in unauth_client.ping().text


def test_ping_auth_with_invalid_account_id(unauth_client):
    """Testing client without proper account id and license key."""
    unauth_client.add_credentials('1234', '4321')
    assert '"authenticated":false' in unauth_client.ping().text



"""Test the add_credential method."""
import pytest
from requests.auth import HTTPBasicAuth
from src.avalara.client import AvataxClient


def test_username_auth(unauth_client):
    """Test passing in username to authorization."""
    unauth_client.add_credentials('joe', '1234')
    assert unauth_client.auth.username == 'joe'


def test_username_auth_2(unauth_client):
    """Test passing in username to authorization."""
    unauth_client.add_credentials('joe', '1234')
    assert unauth_client.auth.password == '1234'


def test_account_id_auth(unauth_client):
    """Test passing in account id to authorization."""
    unauth_client.add_credentials('123321', 'abcdef')
    assert unauth_client.auth.username == '123321'


def test_account_id_auth_2(unauth_client):
    """Test passing in account id to authorization."""
    unauth_client.add_credentials('123321', 'abcdef')
    assert unauth_client.auth.password == 'abcdef'


def test_bearer_token_auth(unauth_client):
    """Test passing in bearer to authorization."""
    unauth_client.add_credentials('123321')
    assert unauth_client.client_header['Authorization'] == 'Bearer 123321'


def test_auth_object(unauth_client):
    """Test auth is HTTPBasicAuth object."""
    unauth_client.add_credentials('joe', '1234')
    assert isinstance(unauth_client.auth, HTTPBasicAuth)


def test_auth_raises_value_error(unauth_client):
    """Test value error is raised with imporper input type."""
    with pytest.raises(ValueError):
        unauth_client.add_credentials(1, False)


def test_auth_able_to_call_and_return_self(unauth_client):
    """Test client can call add_cred method and returns an avatax client obj."""
    assert isinstance(unauth_client.add_credentials(), AvataxClient)


def test_no_auth_if_no_cred_is_added(unauth_client):
    """Test client object auth property is None by default."""
    assert unauth_client.auth is None


def test_if_valid_pair_sandbox_cred_in_place(auth_client):
    """Test if there is a pair of valid Avatax credentials stored in env."""
    assert auth_client is not None



"""Test void transaction method."""


def test_imports(single_transaction, auth_client):
	"""Test import works without errors."""
	trans_code = single_transaction
	assert trans_code is not None


def test_void_changes_transaction_status_to_with_model(single_transaction, auth_client):
	"""Test that running void_transaction voids it."""
	result = auth_client.void_transaction('DEFAULT', single_transaction, {'code':'DocVoided'}).json()
	assert result['status'] == "Cancelled"


def test_void_transaction_with_param(single_transaction_purchase_invoice, auth_client):
	"""Test if we can void a purchase invoice transaction by passing type into param."""
	include = {'documentType': 'PurchaseInvoice'}
	result = auth_client.void_transaction('DEFAULT', single_transaction_purchase_invoice, {'code':'DocVoided'}, include).json()
	assert result['status'] == "Cancelled"




"""Test module for Commit Transaction method."""


def test_commit_transaction_commit_saved_trans(auth_client, single_transaction):
    """Test if we can commit non-commited/saved trans."""
    r = auth_client.commit_transaction('DEFAULT', single_transaction, {'commit': True})
    assert r.json()['status'] == 'Committed'


def test_commit_transaction_wont_commit_if_commit_param_false(auth_client, single_transaction):
    """Test if the transaction will not commited if commit parameter is false."""
    r = auth_client.commit_transaction('DEFAULT', single_transaction, {'commit': False})
    assert r.json()['status'] == 'Saved'




"""Test the client model."""
from src.avalara.client import AvataxClient


def test_client_can_be_created(unauth_client):
    """Test that the client object can be created."""
    assert isinstance(unauth_client, AvataxClient)


def test_client_has_base_url_attribute(unauth_client):
    """Test the client model has attributes available."""
    assert unauth_client.base_url == 'https://sandbox-rest.avatax.com'


def test_client_has_app_name_attribute(unauth_client):
    """Test the client model has attributes available."""
    assert unauth_client.app_name == 'test app'


def test_client_has_app_version_attribute(unauth_client):
    """Test the client model has attributes available."""
    assert unauth_client.app_version == 'ver 0.0'


def test_client_has_machine_name_attribute(unauth_client):
    """Test the client model has attributes available."""
    assert unauth_client.machine_name == 'test machine'


def test_that_client_id_is_created(unauth_client):
    """Test that the client id is created and properly formatted."""
    assert (
        unauth_client.client_id
        == 'test app; ver 0.0; Python SDK; API_VERSION; test machine;'
    )


def test_client_can_obtain_production_url_as_base_url():
    """Test the default option for base url is production url."""
    client = AvataxClient('test app', 'ver 0.0', 'test machine')
    assert client.base_url == 'https://rest.avatax.com'


def test_client_can_obtain_their_own_base_url():
    """Test the client can input a url as the base url."""
    client = AvataxClient(
        'test app', 'ver 0.0', 'test machine', 'https://myurl.com'
    )
    assert client.base_url == 'https://myurl.com'



"""
Ensure all methods needed for new user are functional, including going through
the Transaction State Diagram to verify all methods needed for the lifecycle of 
a single transaction works properly.
"""
import pytest


@pytest.mark.incremental  # allow "test-step" : only continue to next test if previous ones all passes
class TestBasicWorkFlow(object):
	def test_connection(self, auth_client):
		"""Test internet connection of the testing env."""
		assert auth_client.ping().status_code == 200

	def test_ping(self, auth_client):
		"""Test credential is valid."""
		assert '"authenticated":true' in auth_client.ping().text

	def test_check_default_company_exist(self, auth_client, init_comp_model):
		"""Test if the user has default company set up already, create one otherwise."""
		response = auth_client.query_companies()
		assert response.status_code == 200
		if '"companyCode":"DEFAULT"' not in response.text:
			response = auth_client.company_initialize(init_comp_model)
			assert response.status_code == 201
			assert response.json()['isActive'] == True

	def test_create_transaction(self, auth_client, tax_document):
		"""Create a transaction(save it only) using the company created(DEFAULT)."""
		tax_document['commit'] = False
		response = auth_client.create_transaction(tax_document, None)
		assert response.status_code == 201
		assert response.json()['status'] == 'Saved'
		auth_client.trans_code = response.json()['code']

	def test_verfiy_transaction(self, auth_client):
		"""Verify the transaction that has been created."""
		response = auth_client.verify_transaction('DEFAULT', auth_client.trans_code, {})
		assert response.status_code == 200
		assert response.json()['status'] == 'Posted'

	def test_commit_transaction(self, auth_client):
		"""Commit the transaction that has been saved."""
		model = {'commit': True}
		response = auth_client.commit_transaction('DEFAULT', auth_client.trans_code, model)
		assert response.status_code == 200
		assert response.json()['status'] == 'Committed'

	def test_adjust_transaction(self, auth_client, tax_document):
		"""Adjust the transaction just committed."""
		tax_document['commit'] = True
		tax_document['lines'][0]['amount'] = 80
		model = {
					'adjustmentReason': 'PriceAdjusted',
					'adjustmentDescription': 'Price drop before shipping.',
					'newTransaction': tax_document
				}
		response = auth_client.adjust_transaction('DEFAULT', auth_client.trans_code, model)
		assert response.status_code == 200
		assert response.json()['adjustmentReason'] == 'PriceAdjusted'

	def test_void_transaction(self, auth_client):
		"""Void the transaction we just adjusted(still committed status)."""
		model = {'code': 'DocVoided'}
		response = auth_client.void_transaction('DEFAULT', auth_client.trans_code, model)
		assert response.status_code == 200
		assert response.json()['status'] == 'Cancelled'

